# Bineditor - Binary Editor Widgets for Tk
#
# Copyright (c) 2006 OHTSUKA, Yoshio
# This program is free to use, modify, extend at will. The author(s)
# provides no warrantees, guarantees or any responsibility for usage.
# Redistributions in any form must retain this copyright notice.
# ohtsuka.yoshio@gmail.com

package provide Bineditor 1.0
package require Tk

namespace eval bineditor {
	namespace export bineditor
	
	# Initialize OptionDatabase
	proc initOption {className option} {
		upvar $option Option
		foreach name [array names Option] {
			if {[string index $name 0] ne {-}} continue
			set optionName [lindex $Option($name) 0]
			set optionClass [lindex $Option($name) 1]
			set optionValue [lindex $Option($name) 2]
			option add *$className.$optionName $optionValue widgetDefault
		}
		return
	}
	
	# Solve a Alias Option Name
	proc solveAliasOptionName {defaultOption option} {
		upvar $defaultOption Option
		if {[llength $Option($option)] == 1} {
			return $Option($option)
		}
		return $option
	}
	
	#-----------------------------------------------------------
	# Utility
	#-----------------------------------------------------------
	proc convertToNumber {str} {
		if {[catch {expr {$str}} ret]} {
			error "can't convert to numbert \"$str\""
		}
		return $ret
	}

	proc swap {aName bName} {
		upvar $aName a
		upvar $bName b
		set tmp $a
		set a $b
		set b $tmp
	}
	#-----------------------------------------------------------
	# Widget Specific Procs
	#-----------------------------------------------------------
	array set Common {
		className BinEditor
	}

	# Set Widget Default Options
	# if value is not list, that option is an alias of another option.
	# if first char is -, public option, and not, private option
	array set DefaultOption [list \
		-yscrollcommand [list yscrollcommand yScrollCommand {}] \
		-height     [list height Height 10] \
		-insertmode [list insertMode InsertMode overwrite] \
		-background [list background Background black] \
		-foreground [list foreground Foreground white] \
		-bg         -background \
		-fg         -foreground \
		-cursorbg   [list cursorBackground CursorBackground yellow] \
		-cursorfg   [list cursorForeground CursorForeground black] \
		-headerbg   [list headerBackground HeaderBackground black] \
		-headerfg   [list headerForeground HeaderForeground gray60] \
		-selectbg   [list selectBackground SelectBackground #6d66a4] \
		-selectfg   [list selectForeground SelectForeground white] \
		-encoding   [list encoding Encoding ascii] \
		-vikeybind  [list viKeyBind ViKeyBind 0] \
		private {This values is hidden} \
		rootname   {} \
		rootcmd    {} \
	]
	initOption $Common(className) DefaultOption
	

	# Hidden Sub Options
	#option add *$Common(className).Button.background    yellow   widgetDefault
	
	# Widget instance
	namespace eval instance {}
	
	# Create command
	proc bineditor {w args} {
		variable Common
		variable DefaultOption
		
		if {[info command ${w}] ne {}} {
			error "window name \"$w\" already exists in parent"
		}

		# color
		set color(bg)  [lindex $DefaultOption(-background) 2]
		set color(fg)  [lindex $DefaultOption(-foreground) 2]
		set color(hbg) [lindex $DefaultOption(-headerbg) 2]
		set color(hfg) [lindex $DefaultOption(-headerfg) 2]
		set color(inscurbg) [lindex $DefaultOption(-cursorbg) 2]
		set color(inscurfg) [lindex $DefaultOption(-cursorfg) 2]
		set color(selfg) [lindex $DefaultOption(-selectfg) 2]
		set color(selbg) [lindex $DefaultOption(-selectbg) 2]
		
		# font
		switch -exact -- $::tcl_platform(platform) {
			"macintosh" { set npfont Monaco }
			"windows" {
#                set npfont {{MS Gothic} 10} 
				set npfont systemfixed
			}
			"unix" {
				if {[tk windowingsystem] eq "aqua"} {
					set npfont Monaco
				} else {
					set npfont fixed
				}
			}
			default { set npfont fixed }
		}
		# Make widgets
		frame $w -class $Common(className) \
			-background $color(bg)

		label $w.hexHeader \
			-font $npfont \
			-relief flat \
			-width 47 \
			-pady 0 \
			-anchor w \
			-borderwidth 0 \
			-bg $color(hbg) \
			-fg $color(hfg) \
			-text {+0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +A +B +C +D +E +F}

		label $w.asciiHeader \
			-font $npfont \
			-relief flat \
			-width 16 \
			-pady 0 \
			-anchor w \
			-borderwidth 0 \
			-bg $color(hbg) \
			-fg $color(hfg) \
			-text {0123456789ABCDEF}

		text $w.hex \
			-height [lindex $DefaultOption(-height) 2] \
			-undo 1 \
			-autoseparators 0 \
			-font $npfont \
			-relief flat \
			-borderwidth 0\
			-highlightthickness 0\
			-bg $color(bg) \
			-fg $color(fg) \
			-selectbackground $color(selbg) \
			-selectforeground $color(selfg) \
			-insertbackground $color(inscurbg) \
			-width 47 \
			-insertwidth 2 \
			-insertofftime 0 \
			-insertontime  1000

		text $w.offset \
			-height 0 \
			-takefocus 0 \
			-undo 0 \
			-autoseparators 0 \
			-font $npfont \
			-relief flat \
			-borderwidth 0\
			-highlightthickness 0\
			-bg $color(hbg) \
			-fg $color(hfg) \
			-width 7 \
			-insertwidth 2 \
			-state disabled

		text $w.ascii \
			-height 0\
			-undo 0 \
			-autoseparators 0 \
			-wrap none \
			-font $npfont \
			-relief flat \
			-borderwidth 0\
			-highlightthickness 0\
			-bg $color(bg) \
			-fg $color(fg) \
			-insertbackground $color(inscurbg) \
			-width 9 \
			-insertwidth 2 \
			-insertofftime 0 \
			-insertontime  1000

		grid $w.hexHeader   -row 0 -column 1 -sticky ws
		grid $w.asciiHeader -row 0 -column 2 -sticky ws
		grid $w.offset      -row 1 -column 0 -sticky news
		grid $w.hex         -row 1 -column 1 -sticky news  -padx {0 5}
		grid $w.ascii       -row 1 -column 2 -sticky news
		grid rowconfigure    $w 1 -weight 1
		grid columnconfigure $w 3 -weight 1
		grid propagate $w 1

		# Bindings and Tags
		$w.offset configure \
			-selectbackground [$w.offset cget -bg] \
			-selectforeground [$w.offset cget -fg]

		foreach _w [list $w.hex $w.ascii] {
			$_w tag configure sel2 \
				-borderwidth 0 \
				-background $color(selbg) \
				-foreground $color(selfg)

			$_w tag configure insert_focus \
				-borderwidth 0 \
				-relief flat \
				-background $color(inscurbg) \
				-foreground $color(inscurfg)

			$_w tag configure insert_unfocus \
				-borderwidth 2 \
				-relief groove
		}
		$w.hex tag raise insert_focus
		$w.hex tag raise insert_unfocus
		$w.hex tag remove insert_focus 1.0 end
		$w.hex tag remove insert_unfocus 1.0 end
		bindtags $w.hex    [list $w.hex    BinEditor:Hex    . all]
		bindtags $w.ascii  [list $w.ascii  BinEditor:Ascii  . all]
		bindtags $w.offset [list $w.offset BinEditor:Offset . all]
		
		bind $w <Configure> [namespace code {
			update idletask
			Offset:update %W
			Ascii:update %W
			Ascii:updateSelectionFromHex %W
		}]
		
		# create options
		array set [namespace current]::instance::${w} {}
		foreach name [array names DefaultOption] {
			set optionName [lindex $DefaultOption($name) 0]
			set optionClass [lindex $DefaultOption($name) 1]
			set [namespace current]::instance::${w}($name) \
				[option get $w $optionName $optionClass]
		}
		
		# create widget command
		rename ::$w [namespace current]::instance::${w}
		set [namespace current]::instance::${w}(rootname) $w
		set [namespace current]::instance::${w}(rootcmd)  [namespace current]::instance::${w}
		proc ::${w} {command args} \
			"eval [namespace current]::widgetCommand instance::${w} \$command \$args"
		
		
		# destroy proc
		bind $w <Destroy> [namespace code {
			array unset instance::%W
			rename %W {}
		}]

		proc getRootFromTable {w} {return [winfo parent $w]}
		blinkOverwriteCursor $w

		$w configure -insertmode overwrite

		# initialize
		if {[catch {eval widgetCommand instance::$w configure $args} ret]} {
			destroy $w
			error $ret
		}
		
		return $w
	}
	
	# Widget Commands
	proc widgetCommand {w command args} {
		variable DefaultOption
		variable $w
		upvar 0 $w widgetOption
		set cmdname [string map {instance:: {}} $w]

		switch -glob -- $command {
			cget {
				if {[llength $args] != 1} {
					error "wrong # args: should be \"$cmdname cget option\""
				}
				
				if {[array names widgetOption $args] eq {} ||
					[string index $args 0] ne {-}} {
					error "unknown option \"$args\""
				}
				set args [solveAliasOptionName DefaultOption $args]
				return $widgetOption($args)
			}
			
			conf* {
				# show all options
				if {[llength $args] == 0} {
					set result [list]
					foreach name [array names DefaultOption] {
						if {[string index $name 0] ne {-}} continue
						lappend result [concat $name $DefaultOption($name) $widgetOption($name)]
					}
					return $result
				}
				# show a option
				if {[llength $args] == 1} {
					if {[array names DefaultOption $args] eq {} ||
						[string index $args 0] ne {-}} {
						error "unknown option \"$args\""
					}
					foreach name [array names DefaultOption $args] {
						set name [solveAliasOptionName DefaultOption $name]
						return [concat $name $DefaultOption($name) $widgetOption($name)]
					}
				}
				#check all and set options
				foreach {key value} $args {
					if {[array names DefaultOption $key] eq {} ||
						[string index $key 0] ne {-}} {
						error "unknown option \"$key\""
					}
				}
				foreach {key value} $args {
					set key [solveAliasOptionName DefaultOption $key]
					switch -- $key {
						-yscrollcommand {
							$widgetOption(rootname).hex configure -yscrollcommand $value
							set widgetOption($key) $value
						}
						-height {
							$widgetOption(rootname).hex configure -height $value
							set widgetOption($key) $value
						}
						-insertmode {
							switch -exact -- $value {
								"insert"    -
								"overwrite" {set widgetOption(-insertmode) $value}
								default {
								error "bad insert mode \"$value\": must be insert or overwrite"
								}
							}
							switch -exact -- $value {
								"insert" {
									$widgetOption(rootname).hex configure -insertwidth 2
									$widgetOption(rootname).ascii configure -insertwidth 2
								}
								"overwrite" {
									$widgetOption(rootname).hex configure -insertwidth 0
									$widgetOption(rootname).ascii configure -insertwidth 0
								}
							}
						}
						-background {
							$widgetOption(rootcmd)        configure -bg $value
							$widgetOption(rootname).hex   configure -bg $value
							$widgetOption(rootname).ascii configure -bg $value
							set widgetOption($key) $value
						}
						-foreground {
							$widgetOption(rootname).hex   configure -fg $value
							$widgetOption(rootname).ascii configure -fg $value
							set widgetOption($key) $value
						}
						-cursorbg {
							$widgetOption(rootname).hex   tag configure insert -background $value
							$widgetOption(rootname).ascii tag configure insert -background $value
							$widgetOption(rootname).hex   configure -insertbackground $value
							$widgetOption(rootname).ascii configure -insertbackground $value
							set widgetOption($key) $value
						}
						-cursorfg {
							$widgetOption(rootname).hex   tag configure insert -foreground $value
							$widgetOption(rootname).ascii tag configure insert -foreground $value
							set widgetOption($key) $value
						}
						-headerbg {
							$widgetOption(rootname).offset      configure -bg $value
							$widgetOption(rootname).asciiHeader configure -bg $value
							$widgetOption(rootname).hexHeader   configure -bg $value
							set widgetOption($key) $value
						}
						-headerfg {
							$widgetOption(rootname).offset      configure -fg $value
							$widgetOption(rootname).asciiHeader configure -fg $value
							$widgetOption(rootname).hexHeader   configure -fg $value
							set widgetOption($key) $value
						}
						-selectbg {
							$widgetOption(rootname).hex   tag configure sel2 -background $value
							$widgetOption(rootname).ascii tag configure sel2 -background $value
							set widgetOption($key) $value
						}
						-selectfg {
							$widgetOption(rootname).hex   tag configure sel2 -foreground $value
							$widgetOption(rootname).ascii tag configure sel2 -foreground $value
							set widgetOption($key) $value
						}
						-encoding {
							if {[lsearch [encoding names] $value] < 0} {
								error "unknown encoding \"$value\""
							}
							set widgetOption($key) $value
						}
						-vikeybind {
							set hexw $widgetOption(rootname).hex 
							if {[string is boolean $value] && $value} {
								bind $hexw <h> "event generate $hexw <Left>"
								bind $hexw <j> "event generate $hexw <Down>"
								bind $hexw <k> "event generate $hexw <Up>"
								bind $hexw <l> "event generate $hexw <Right>"
								bind $hexw <u> "event generate $hexw <Control-z>"
							} else {
								bind $hexw <h> {}
								bind $hexw <j> {}
								bind $hexw <k> {}
								bind $hexw <l> {}
								bind $hexw <u> {}
							}
							set widgetOption($key) $value
						}
						default {}
					}
				}
				return
			}
			yview {
				set root $widgetOption(rootname)
				if {[llength $args] == 0} {
					return [$root.hex yview]
				}
				eval [concat $root.hex yview $args]
				Offset:update $root
				Ascii:update $root
				Ascii:updateSelectionFromHex $root
			}
			set {
				if {[llength $args] != 1} {
					error "wrong # args: should be \"$cmdname set data\""
				}
				set root $widgetOption(rootname)
				set data [binToHex [lindex $args 0]]
				set hex $root.hex
				$hex delete 1.0 end
				$hex insert 1.0 $data
				::update
				Offset:update $root
				Ascii:update $root
				$hex edit reset
			}
			insert {
				if {[llength $args] != 2} {
					error "wrong # args: should be \"$cmdname insert offset data\""
				}
				set offset [lindex $args 0]
				if {$offset < 0 || [catch {expr {$offset}}]} {
					error "selection set error:\"$offset\" invalid argument"
				}
				set index [Hex:convertOffsetToIndex $offset]
				set bin [lindex $args 1]
				set hex $widgetOption(rootname).hex
				Hex:insertBin $hex "insert" 1 $index $bin
			}
			overwrite {
				if {[llength $args] != 2} {
					error "wrong # args: should be \"$cmdname overwrite offset data\""
				}
				set offset [lindex $args 0]
				if {$offset < 0 || [catch {expr {$offset}}]} {
					error "selection set error:\"$offset\" invalid argument"
				}
				set index [Hex:convertOffsetToIndex $offset]
				set bin [lindex $args 1]
				set hex $widgetOption(rootname).hex
				Hex:insertBin $hex "overwrite" 1 $index $bin
			}
			get {
				if {[llength $args] == 0} {
					set hseq [string map {\n {} { } {}} [$widgetOption(rootname).hex get 0.0 end]]
					return [binary format H* $hseq]
				}
				if {[llength $args] == 1} {
					set offset [convertToNumber [lindex $args 0]]
					set index [Hex:convertOffsetToIndex $offset]
					set hseq [string map {\n {} { } {}} \
						[$widgetOption(rootname).hex get $index "$index +3 c"]]
					return [binary format H* $hseq]
				}
				if {[llength $args] == 2} {
					set offset1 [convertToNumber [lindex $args 0]]
					set offset2 [convertToNumber [lindex $args 1]]
					set index1 [Hex:convertOffsetToIndex $offset1]
					set index2 [Hex:convertOffsetToIndex $offset2]
					set hseq [string map {\n {} { } {}} \
						[$widgetOption(rootname).hex get $index1 "$index2 +3 c"]]
					return [binary format H* $hseq]
				}
				error "wrong # args: should be \"$cmdname get ?offset1? ?offset2?\""
			}
			clear  {
				if {[llength $args] != 0} {
					error "wrong # args: should be \"$cmdname clear\""
				}
				$widgetOption(rootname).hex delete 1.0 end
				Ascii:update $widgetOption(rootname)
				Offset:update $widgetOption(rootname)
				$widgetOption(rootname).hex edit reset
				return
			}
			offset {
				if {[llength $args] != 1} {
					error "wrong # args: should be \"$cmdname offset mark\""
				}
				set hw $widgetOption(rootname).hex
				switch -exact -- [lindex $args 0] {
					cursor    {set index [$hw index insert]}
					last      {set index [$hw index end-1c]}
					sel.first {if {[catch {$hw index sel2.first} index]} {return}}
					sel.last  {if {[catch {$hw index sel2.last}  index]} {return}}
					default {error "unknown mark : must be cursor, last, sel.first or sel.last"}
				}
				return [Hex:convertIndexToOffset $index]
			}
			movecursor {
				if {[llength $args] != 1} {
					error "wrong # args: should be \"$cmdname movecursor offset\""
				}
				set index [Hex:convertOffsetToIndex [lindex $args 0]]
				set hw $widgetOption(rootname).hex
				Hex:setCursor $hw [Hex:getCursor $hw $index] 0
			}
			see {
				if {[llength $args] != 1} {
					error "wrong # args: should be \"$cmdname see offset\""
				}
				set index [Hex:convertOffsetToIndex [lindex $args 0]]
				set hw $widgetOption(rootname).hex
				Hex:see $hw [Hex:getCursor $hw $index] 1
			}
			selection {
				set root $widgetOption(rootname)
				set hw $widgetOption(rootname).hex
				
				switch -exact -- [lindex $args 0] {
					set {
						if {[llength $args] != 3} {
							error "wrong # args: should be \"$cmdname selection set offset1 offset2\""
						}
						set offset1 [lindex $args 1]
						set offset2 [lindex $args 2]
						if {$offset1 < 0} {error "selection set error:\"$offset1\" invalid argument"}
						if {$offset2 < 0} {error "selection set error:\"$offset2\" invalid argument"}
						if {$offset1 > $offset2} {swap offset1 offset2}
						set index1 [Hex:convertOffsetToIndex $offset1]
						set index2 [Hex:convertOffsetToIndex $offset2]
						Hex:clearSelection $hw
						Hex:setSelectionAnchor $hw [Hex:getCursor $hw $index1]
						Hex:setSelection $hw [Hex:getCursor $hw $index2]
						Ascii:updateSelectionFromHex $root
						return
					}
					get {
						if {[llength [$hw tag ranges sel2]] < 1} {return [list]}
						return [list \
							[Hex:convertIndexToOffset [$hw index sel2.first]] \
							[Hex:convertIndexToOffset [$hw index sel2.last]]]
					}
				}
				error "wrong # args: should be \"$cmdname selection set or get\""
			}
			default {
				# unknown command
				set cmds {cget configure yview set get insert overwrite clear offset movecursor see selection}
				set cmds [lsort $cmds]
				set str "ambiguous command \"$command\": must be "
				append str [join [lrange $cmds 0 end-1] ", "]
				append str " or " [lindex $cmds end]
				error $str 
			}
		}
	}
	
	#=======================================================================
	# Commom
	#=======================================================================
	proc binToHex {bin {offset 0}} {
		set offset [expr {16 - ($offset % 16)}]
		set len [string bytelength $bin]
		set data ""

		if {$offset > 0} {
			set nset [binary scan $bin [string repeat H2 $offset] \
			h0 h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 h11 h12 h13 h14 h15]
			set tlist [list]
			for {set i 0} {$i < $nset} {incr i} {
				lappend tlist [set h${i}]
			}
			append data [join $tlist " "]\n
		}

		set H64 [string repeat H2 64]
		for {set i $offset} {$i < $len} {incr i 64} {
			set nset [binary scan $bin @${i}${H64} \
			h0 h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 h11 h12 h13 h14 h15 \
			h16 h17 h18 h19 h20 h21 h22 h23 h24 h25 h26 h27 h28 h29 h30 h31 \
			h32 h33 h34 h35 h36 h37 h38 h39 h40 h41 h42 h43 h44 h45 h46 h47 \
			h48 h49 h50 h51 h52 h53 h54 h55 h56 h57 h58 h59 h60 h61 h62 h63]
			
			if {$nset != 64} {
				set line [list]
				for {set j 0} {$j < $nset} {} {
					lappend line [set h${j}]
					incr j
					
					if {$j % 16 == 0} {
						lappend line \n
					} else {
						lappend line { }
					}
				}
				append data [join $line {}]
			} else {
				#open time:6.441 sec
				append data \
				"$h0 $h1 $h2 $h3 $h4 $h5 $h6 $h7 $h8 $h9 $h10 $h11 $h12 $h13 $h14 $h15\n"\
				"$h16 $h17 $h18 $h19 $h20 $h21 $h22 $h23 $h24 $h25 $h26 $h27 $h28 $h29 $h30 $h31\n"\
				"$h32 $h33 $h34 $h35 $h36 $h37 $h38 $h39 $h40 $h41 $h42 $h43 $h44 $h45 $h46 $h47\n"\
				"$h48 $h49 $h50 $h51 $h52 $h53 $h54 $h55 $h56 $h57 $h58 $h59 $h60 $h61 $h62 $h63\n"
			}
		}
		return [string toupper $data]
	}

	proc Common:setCursor {w index {focus 1}} {
		set root [getRootFromTable $w]
		$w tag remove insert_focus 1.0 end
		$w tag remove insert_unfocus 1.0 end
		$w mark set insert $index
		$w mark set anchor $index
		set mode [$root cget -insertmode]
		if {$mode ne "insert"} {
			if {$focus} {
				$w tag add insert_focus insert
			} else {
				$w tag add insert_unfocus insert
			}
		}
	}

	proc blinkOverwriteCursor {root} {
		set interval 500
		set blink [expr {([clock clicks -milliseconds]/$interval) % 2}]
		set mode [$root cget -insertmode]
		if {$mode eq "overwrite"} {
			foreach c {hex ascii} {
				if {[focus] eq "$root.$c"} {
					$root.$c tag remove insert_unfocus 1.0 end
					if {$blink && [lsearch [$root.$c mark names] insert_hide] < 0} {
						$root.$c tag add insert_focus insert
					} else {
						$root.$c tag remove insert_focus 1.0 end
					}
				} else {
					$root.$c tag remove insert_focus 1.0 end
					if {$blink && [lsearch [$root.$c mark names] insert_hide] < 0} {
						$root.$c tag add insert_unfocus insert
					} else {
						$root.$c tag remove insert_unfocus 1.0 end
					}
				}
			}
		} else {
			$root.hex   tag remove insert_focus 1.0 end
			$root.ascii tag remove insert_focus 1.0 end
			$root.hex   tag remove insert_unfocus 1.0 end
			$root.ascii tag remove insert_unfocus 1.0 end
			if {[focus] eq "$root.ascii"} {
				if {$blink && [lsearch [$root.ascii mark names] insert_hide] < 0} {
					$root.ascii configure -insertwidth 2
				} else {
					$root.ascii configure -insertwidth 0
				}
			}
			if {[focus] eq "$root.hex"} {
				if {$blink} {
					$root.hex configure -insertwidth 2
				} else {
					$root.hex configure -insertwidth 0
				}
			}
			::update idletask

		}
		after $interval [namespace code [info level [info level]]]
	}

	proc changeMode {w} {
		set root [getRootFromTable $w]
		set mode [$root cget -insertmode]
		if {$mode eq "insert"} {
			$root configure -insertmode "overwrite"
		} else {
			$root configure -insertmode "insert"
		}
	}

	proc incrIndex {index line char} {
		scan $index %d.%d linecount charcount
		return [incr linecount $line].[incr charcount $char]
	}

	proc getTextLine {index} {
		scan $index %d.%d linecount charcount
		return $linecount
	}

	proc getTextChar {index} {
		scan $index %d.%d linecount charcount
		return $charcount
	}

	proc isOffscreen {wt index} {
		set bbox [$wt bbox $index]
		if {[llength $bbox] < 1} { return 1 }
		foreach {x y w h} $bbox {break}
		set ch [font metrics [$wt cget -font] -linespace]
		return [expr {$h != $ch}]
	}
	
	#=======================================================================
	# BinEditor:Hex
	#=======================================================================
	bind BinEditor:Hex <1> [namespace code {
		Hex:setSelectionAnchor %W [Hex:getCursor %W @%x,%y]
		Hex:setCursor %W [Hex:getCursor %W current] 1
	}]
	bind BinEditor:Hex <Button1-Motion> [namespace code {
		Hex:setSelection %W [Hex:getCursor %W @%x,%y]
		Ascii:updateSelectionFromHex [getRootFromTable %W]
	}]
	bind BinEditor:Hex <Shift-1> [namespace code {
		Hex:setSelection %W [Hex:getCursor %W @%x,%y]
		Ascii:updateSelectionFromHex [getRootFromTable %W]
	}]
	bind BinEditor:Hex <MouseWheel> [namespace code {
		Hex:eventMouseWheel %W %D
	}]
	bind BinEditor:Hex <Right>     [namespace code {Hex:eventRight %W 1}]
	bind BinEditor:Hex <Left>      [namespace code {Hex:eventLeft  %W 1}]
	bind BinEditor:Hex <Up>        [namespace code {Hex:eventUp    %W 1}]
	bind BinEditor:Hex <Down>      [namespace code {Hex:eventDown  %W 1}]
	bind BinEditor:Hex <Insert>    [namespace code {changeMode %W}]
	bind BinEditor:Hex <Delete>    [namespace code {Hex:eventDelete    %W 1}]
	bind BinEditor:Hex <BackSpace> [namespace code {Hex:eventBackSpace %W 1}]
	bind BinEditor:Hex <Control-z> [namespace code {Hex:eventControl-z %W 1}]
	bind BinEditor:Hex <KeyPress>  [namespace code {Hex:eventKeyPress %W %A 1}]
	bind BinEditor:Hex <<Cut>>   [namespace code {Hex:cut   %W}]
	bind BinEditor:Hex <<Copy>>  [namespace code {Hex:copy  %W}]
	bind BinEditor:Hex <<Paste>> [namespace code {Hex:paste %W}]
	bind BinEditor:Hex <<Clear>> [namespace code {Hex:clear %W}]
	
	proc Hex:eventMouseWheel {w d} {
		$w yview scroll [expr {$d < 0 ? 4 : -4}] units
		set root [getRootFromTable $w]
		Offset:update $root
		Ascii:update $root
		Ascii:updateSelectionFromHex $root
	}
	proc Hex:eventDown {w focus} {
		Hex:setCursor $w [Hex:getUpDownLineIndex $w down {insert wordstart}] $focus
	}
	proc Hex:eventUp {w focus} {
		Hex:setCursor $w [Hex:getUpDownLineIndex $w up {insert wordstart}] $focus
	}
	proc Hex:eventLeft {w focus} {
		Hex:setCursor $w [Hex:getCursor $w "insert -3 c"] $focus
	}
	proc Hex:eventRight {w focus} {
		Hex:setCursor $w [Hex:getCursor $w "insert +3 c"] $focus
	}
	proc Hex:eventDelete {w focus} {
		if {[llength [$w tag ranges sel2]] > 0} {
			Hex:delete $w 1 sel2.first sel2.last
		} else {
			Hex:delete $w 1 insert
		}
		if {$focus} {focus $w}
		event generate [getRootFromTable $w] <<Cursor>>
	}
	proc Hex:eventBackSpace {w focus} {
		if {[llength [$w tag ranges sel2]] > 0} {
			Hex:delete $w 1 sel2.first sel2.last
		} else {
			if {[$w compare insert > 1.0]} {
				Common:setCursor $w [Hex:getCursor $w "insert-3c"]
				Hex:delete $w 1 insert
				$w edit separator
			}
		}
		if {$focus} {focus $w}
		event generate [getRootFromTable $w] <<Cursor>>
	}
	proc Hex:eventControl-z {w focus} {
		if {[catch {$w edit undo}]} {bell; return}
		if {[string is space [$w get insert]]} {
			Hex:setCursor $w {insert -1 c wordstart} $focus
		} else {
			Hex:setCursor $w {insert wordstart} $focus
		}
	}
	proc Hex:eventKeyPress {w s focus} {
		set root [getRootFromTable $w]
		set s [string toupper $s]
		if {$s eq "" || [$w cget -state] eq "disabled" ||
			[string first $s {0123456789ABCDEF}] < 0} {
			return
		}
		if {[string is space [$w get insert-1c]]} {
			$w edit separator
		}
		if {[$w compare insert >= end-1c]} {
			set mode insert
		} else {
			set mode [$root cget -insertmode]
		}
		switch -- $mode {
			"insert" {
				if {[string is space [$w get "insert -1 c"]] ||
					[$w compare insert <= 1.0]} {
					set tmpcur [$w index insert]
					$w insert insert "${s}0 "
					
					Hex:rebuildLines $w insert
					Hex:setCursor $w "$tmpcur +1 c" $focus
				} else {
					$w delete insert
					$w insert insert $s
					if {[string is space [$w get insert]]} {
						$w edit separator
						Hex:setCursor $w {insert +1 c} $focus
					}
				}
			}
			"overwrite" {
				$w delete insert
				$w insert insert $s
				if {[string is space [$w get insert]]} {
					Hex:setCursor $w {insert +1 c} $focus
					$w edit separator
				}
			}
		}
	}
	proc Hex:cut {w} {
		if {[$w tag ranges sel2] eq ""} {return}
		Hex:copy $w
		Hex:delete $w 1 sel2.first sel2.last
		event generate [getRootFromTable $w] <<Cursor>>
	}
	proc Hex:copy {w} {
		if {[$w tag ranges sel2] eq ""} {return}
		clipboard clear
		clipboard append [$w get sel2.first sel2.last]

	}
	proc Hex:paste {w} {
		set hstr [clipboard get]
		set hstr [string map {\n {} \t {} - {} , {} . {} { } {}} $hstr]
		if {[catch {binary format H* $hstr} bin]} {
			bell; return
		}
		
		set root [getRootFromTable $w]
		set mode [$root cget -insertmode]
		if {$mode eq "insert"} {
			if {[llength [$w tag ranges sel2]] > 0} {
				Hex:delete $w 0 sel.first sel.last
			}
		}
		Hex:insertBin $w [$root cget -insertmode] 1 [$w index insert] $bin
		event generate [getRootFromTable $w] <<Cursor>>
	}
	proc Hex:clear {w} {
		Hex:delete $w 1 1.0 [$w index end]
		$w edit separator
		event generate [getRootFromTable $w] <<Cursor>>
	}

	proc Hex:getCursor {w index} {
		if {[string is space [$w get $index]]} {
			return [$w index "$index +1 char"]
		} else {
			return [$w index "$index wordstart"]
		}
	}
	
	proc Hex:getUpDownLineIndex {w dir index} {
		set maxline [getTextLine [$w index "end-1c"]]
		set curline [getTextLine [$w index $index]]
		if {$dir eq "up"} {
			if {$curline <= 1} {return $index}
			return [$w index "$index - 1 lines"]
		} else {
			if {$curline >= $maxline} {return $index}
			return [$w index "$index + 1 lines"]
		}
	}
	
	proc Hex:setCursor {w index {focus 1}} {
		Common:setCursor $w $index $focus
		$w see $index
		$w tag remove sel2 1.0 end
		Hex:setSelectionAnchor $w insert
		if {$focus} { focus $w }
		set root [getRootFromTable $w]
		Offset:update $root
		Ascii:update $root
		event generate [getRootFromTable $w] <<Cursor>>
	}
	
	proc Hex:see {w index {update 1}} {
		set root [getRootFromTable $w]
		$w see $index
		if {$update} {
			Offset:update $root
			Ascii:update $root
		}
	}
	
	proc Hex:setSelection {w index} {
		$w tag remove sel2 1.0 end
		Common:setCursor $w $index [expr {[focus] eq $w}]
		
		if {[$w compare selanchor <= insert]} {
			$w tag add sel2 selanchor insert
		} else {
			$w tag add sel2 insert selanchor 
		}
		event generate [getRootFromTable $w] <<Cursor>>
	}

	proc Hex:clearSelection {w} {
		$w tag remove sel2 1.0 end
	}

	proc Hex:setSelectionAnchor {w index} {
		$w mark set selanchor $index
	}

	proc Hex:rebuildLines {w index} {
		set index [$w index $index]
		set str [$w get "$index linestart" end]

		if {[getTextLine $index] == 1} {
			set nstr ""
		} else {
			set nstr \n
		}
		set len [llength $str]
		for {set i 0} {$i < $len} {incr i 16} {
			append nstr [lrange $str $i [expr {$i + 15}]]\n
		}
		$w delete "$index linestart" end
		$w insert end [string trimright $nstr \n]
		if {[getTextChar [$w index "end -1 c"]] == 47} {
			$w insert end \n
		} else {
			$w insert end " "
		}
	}
	
	# insert binarydata to index. this is called from other procs.
	proc Hex:insertBin {w mode undo index bin} {
		set root [getRootFromTable $w]

		$w mark set insert $index

		switch -- $mode {
			"insert" {
				set tmpcur [$w index insert]
				set hex [binToHex $bin]
				$w insert insert $hex
				Hex:rebuildLines $w $tmpcur
				set next [expr {[llength $hex] * 3}]
				set nextindex [$w index "$tmpcur + $next c"]
				$w mark set insert    $nextindex
				$w mark set selanchor $nextindex
			}
			"overwrite" {
				binary scan $bin H* hex
				set hex [string toupper $hex]
				foreach {a b} [split $hex {}] {
					if {[$w compare insert >= end-1c]} {
						$w insert insert "$a$b"
						if {[getTextChar [$w index insert]] == 47} {
							$w insert insert "\n"
						} else {
							$w insert insert " "
						}
					} else {
						$w delete insert {insert +2 c}
						$w insert insert "$a$b"
						$w mark set insert    {insert +1 c}
						$w mark set selanchor {insert +1 c}
					}
				}
			}
		}
		$w tag remove sel2 1.0 end
		Hex:see $w [$w index insert] 0
		if {$undo} {$w edit separator}
		Offset:update $root
		Ascii:update $root
	}
	
	proc Hex:delete {w undo index0 {index1 {}}} {
		set root [getRootFromTable $w]
		if {$index1 eq ""} {
			set tmpcur [$w index insert]
			$w delete $index0 "$index0 +3 c"
		} else {
			if {[$w compare $index0 > $index1]} {
				swap index0 index1
			}
			set tmpcur [$w index $index0]
			$w delete $index0 $index1
		}
		if {[$w compare 1.0 == "end-1c"]} {
			$root clear
		} else {
			Hex:rebuildLines $w $tmpcur
		}
		$w mark set insert $tmpcur
		$w mark set selanchor insert
		if {$undo} {$w edit separator}
		Offset:update $root
		Ascii:update $root
	}
	
	proc Hex:getViewRange {root} {
		set hex $root.hex
		set x [winfo x $hex]
		set y [winfo y $hex]
		set h [winfo h $hex]
		set stindex [$hex index @$x,$y]
		set edindex [$hex index @$x,[expr {$y + $h}]]
		scan $stindex "%d.%d" stline stchar
		scan $edindex "%d.%d" edline edchar
		return [list $stline $edline]
	}
	
	proc Hex:convertToAsciiIndex {w index} {
		set root [getRootFromTable $w]
		foreach {stline edline} [Hex:getViewRange $root] {break}
		set index [$w index $index]
		set line [getTextLine $index]
		set char [getTextChar $index]
		return [expr {$line - $stline + 1}].[expr {$char / 3}]
	}

	proc Hex:convertOffsetToIndex {offset} {
		if {[catch {expr {$offset}}]} {
			error "\"$offset\" is not number"
		}
		set line [expr {($offset / 16) +1}]
		set char [expr {($offset % 16) *3}]
		return ${line}.${char}
	}
	proc Hex:convertIndexToOffset {index} {
		set line [getTextLine $index]
		set char [getTextChar $index]
		return [expr {($line -1) * 16 + $char/3}]
	}
	#=======================================================================
	# BinEditor:Ascii
	#=======================================================================
	bind BinEditor:Ascii <1> [namespace code {
		Ascii:setCursor %W [Ascii:getCursor %W current] 1
		%W mark set selanchor [Ascii:getCursor %W @%x,%y]
		%W mark unset insert_hide
		Hex:setSelectionAnchor [getRootFromTable %W].hex \
		[Ascii:convertToHexIndex %W selanchor]
		[getRootFromTable %W].hex mark set insert [Ascii:convertToHexIndex %W insert]
		Hex:clearSelection [getRootFromTable %W].hex
		event generate [getRootFromTable %W] <<Cursor>>
	}]
	bind BinEditor:Ascii <Button1-Motion> [namespace code {
		Ascii:setCursor %W [Ascii:getCursor %W @%x,%y] 1
		Hex:setSelection [getRootFromTable %W].hex [Ascii:convertToHexIndex %W @%x,%y]
		Ascii:updateSelectionFromHex [getRootFromTable %W]
		focus %W
	}]

	bind BinEditor:Ascii <MouseWheel> [namespace code {
		Hex:eventMouseWheel [getRootFromTable %W].hex %D
	}]

	bind BinEditor:Ascii <Shift-1> [namespace code {
		Hex:setSelection [getRootFromTable %W].hex [Ascii:convertToHexIndex %W @%x,%y]
		Ascii:updateSelectionFromHex [getRootFromTable %W]
		focus %W
	}]
	bind BinEditor:Ascii <Right> [namespace code {
		Hex:eventRight [getRootFromTable %W].hex 0
		Ascii:setCursor %W [%W index insert] 1
	}]
	bind BinEditor:Ascii <Left> [namespace code {
		Hex:eventLeft [getRootFromTable %W].hex 0
		Ascii:setCursor %W [%W index insert] 1
	}]
	bind BinEditor:Ascii <Up> [namespace code {
		Hex:eventUp [getRootFromTable %W].hex 0
		Ascii:setCursor %W [%W index insert] 1
	}]
	bind BinEditor:Ascii <Down> [namespace code {
		Hex:eventDown [getRootFromTable %W].hex 0
		Ascii:setCursor %W [%W index insert] 1
	}]
	bind BinEditor:Ascii <Insert> [namespace code {
		changeMode %W
	}]
	bind BinEditor:Ascii <Delete> [namespace code {
		Hex:eventDelete [getRootFromTable %W].hex 0
	}]
	bind BinEditor:Ascii <BackSpace> [namespace code {
		Hex:eventBackSpace [getRootFromTable %W].hex 0
	}]
	bind BinEditor:Ascii <Control-z> [namespace code {
		Hex:eventControl-z [getRootFromTable %W].hex 0
	}]
	bind BinEditor:Ascii <KeyPress> [namespace code {
		Ascii:eventKeyPress %W %A 1
	}]

	bind BinEditor:Ascii <<Cut>>   [namespace code {Hex:cut   [getRootFromTable %W].hex}]
	bind BinEditor:Ascii <<Copy>>  [namespace code {Hex:copy  [getRootFromTable %W].hex}]
	bind BinEditor:Ascii <<Paste>> [namespace code {Hex:paste [getRootFromTable %W].hex}]
	bind BinEditor:Ascii <<Clear>> [namespace code {Hex:clear [getRootFromTable %W].hex}]

	proc Ascii:eventKeyPress {w a focus} {
		if {$a eq ""||$a eq "\t"} {return}
		set root [getRootFromTable $w]
		set hex $root.hex
		set bin [encoding convertto [$root cget -encoding] $a]
		Hex:insertBin $hex [$root cget -insertmode] 1 insert $bin
		if {$focus} {focus $w}
		event generate [getRootFromTable $w] <<Cursor>>
	}

	proc Ascii:setCursor {w index focus} {
		set root [getRootFromTable $w]
		Common:setCursor $w $index
		$w tag remove sel2 1.0 end
		$root.hex mark set insert [Ascii:convertToHexIndex $w insert]
		if {$focus} {focus $w}
	}

	proc Ascii:getCursor {w index} {
		return [$w index $index]
	}

	proc Ascii:update {root} {
		foreach {stline edline} [Hex:getViewRange $root] {break}
		set hex   $root.hex
		set ascii $root.ascii

		set str ""
		set lines [$hex get $stline.0 $edline.end]
		
		foreach line [split $lines \n] {
			set newline ""
			foreach byte $line {
				set num [scan $byte %02x]
				if {$num < 0x20 || $num > 0x7E} {
					append newline 2E
				} else {
					append newline $byte
				}
			}
			append str [binary format H* $newline]\n
		}
		
		# set str
		$ascii tag remove sel2 1.0 end
		$ascii delete 1.0 end
		$ascii insert end [string range $str 0 end-1]
		
		# set mark
		set index [Hex:convertToAsciiIndex $hex insert]
		set line [getTextLine $index]
		if {$line < 1 || $line > ($edline - $stline + 1)} {
			$ascii mark set insert_hide 1.0
			$ascii tag remove insert_focus 1.0 end
			$ascii tag remove insert_unfocus 1.0 end
		} else {
			$ascii mark unset insert_hide
			$ascii mark set insert $index
			Common:setCursor $ascii insert [expr {$ascii eq [focus]}]
		}
	}

	proc Ascii:updateSelectionFromHex {root} {
		foreach {stline edline} [Hex:getViewRange $root] {break}
		set hex   $root.hex
		set ascii $root.ascii
		catch {
			$ascii tag remove sel2 1.0 end
			set st [Hex:convertToAsciiIndex $hex sel2.first]
			set ed [Hex:convertToAsciiIndex $hex sel2.last]
			$ascii tag add sel2 $st $ed
		}
		$ascii mark set insert [Hex:convertToAsciiIndex $hex insert]
		Common:setCursor $ascii [Hex:convertToAsciiIndex $hex insert] 0
	}
	
	proc Ascii:convertToHexIndex {w index} {
		set root [getRootFromTable $w]
		set hex $root.hex
		foreach {stline edline} [Hex:getViewRange $root] {break}
		scan [$w index $index] "%d.%d" line char
		return [expr {$stline + $line -1}].[expr {$char * 3}]
	}

	#=======================================================================
	# BinEditor:Offset
	#=======================================================================

	proc Offset:update {root} {
		foreach {stline edline} [Hex:getViewRange $root] {break}
		set offset $root.offset
		set num {}
		for {set i $stline} {$i <= $edline} {incr i} {
			append num [format %06X [expr {($i - 1) * 16}]] \n
		}
		$offset configure -state normal
		$offset delete 1.0 end
		$offset insert 1.0 $num
		$offset configure -state disabled
	}
}
